/*============================================================================
  File:     02_EssentialInfo.sql

  SQL Server Versions:  SQL Server Versions: 2016, 2017
------------------------------------------------------------------------------
  Written by Erin Stellato, SQLskills.com
  
  (c) 2019, SQLskills.com. All rights reserved.

  For more scripts and sample code, check out 
    http://www.SQLskills.com

  You may alter this code for your own *non-commercial* purposes. You may
  republish altered code as long as you include this copyright and give due
  credit, but you must obtain prior permission before blogging this code.
  
  THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
============================================================================*/

/*
	Find query parameters and data type
	Enable showplan
*/
USE [WideWorldImporters];
GO

DROP PROCEDURE IF EXISTS [Sales].[usp_CustomerTransactionInfo];
GO

CREATE PROCEDURE [Sales].[usp_CustomerTransactionInfo]
	@CustomerID INT
AS	

	SELECT [CustomerID], SUM([AmountExcludingTax])
	FROM [Sales].[CustomerTransactions]
	WHERE [CustomerID] = @CustomerID
	GROUP BY [CustomerID];
GO

/*
	Execute individually with different input parameters
	Note cost, I/O, how to find input parameters in actual plan
*/
EXEC [Sales].[usp_CustomerTransactionInfo] 1050;
GO

EXEC [Sales].[usp_CustomerTransactionInfo] 401;
GO


/*
	Retrieve plan from cache
*/
SET TRANSACTION ISOLATION LEVEL READ UNCOMMITTED;
GO
SELECT 
	[qs].execution_count, 
	[s].[text], 
	[qs].[query_hash], 
	[qs].[query_plan_hash], 
	[cp].[size_in_bytes]/1024 AS [PlanSizeKB], 
	[qp].[query_plan], 
	[qs].[plan_handle]
FROM sys.dm_exec_query_stats AS [qs]
CROSS APPLY sys.dm_exec_query_plan ([qs].[plan_handle]) AS [qp]
CROSS APPLY sys.dm_exec_sql_text([qs].[plan_handle]) AS [s]
INNER JOIN sys.dm_exec_cached_plans AS [cp] 
	ON [qs].[plan_handle] = [cp].[plan_handle]
WHERE [s].[text] LIKE '%CustomerTransactionInfo%';
GO
SET TRANSACTION ISOLATION LEVEL READ COMMITTED;
GO

/*
	clear cache and re-run
*/
ALTER DATABASE SCOPED CONFIGURATION CLEAR PROCEDURE_CACHE;
GO

EXEC [Sales].[usp_CustomerTransactionInfo] 401;
GO


/*
	Queries are sensitive to input parameters and have different costs...
	Start with cleared cache
*/
ALTER DATABASE SCOPED CONFIGURATION CLEAR PROCEDURE_CACHE;
GO


/*
	1) Relative Batch costs (operations that are invisible to the optimizer (MSTVF) can skew this output)
		Is it "best" to start with the plan that has a higher cost?
	2) Within statement
		Often prioritize where to tune based on cost
	3) Subtree cost (cost of current operator and descendants)
	4) Not every operator has CPU and I/O cost

	** Execute two SPs together, THEN execute two SELECTs together **
*/
EXEC [Sales].[usp_CustomerTransactionInfo] 1050;

EXEC [Sales].[usp_CustomerTransactionInfo] 401;
GO

/*
	Stop...compare plans in QS 
*/

SELECT 
	[StockItemID], 
	[Quantity]
FROM [Sales].[OrderLines]
WHERE [StockItemID] = 223
ORDER BY [StockItemID]

SELECT 
	[o].[OrderID], 
	[ol].[OrderLineID],
	[o].[OrderDate], 
	[o].[CustomerID],
	[ol].[StockItemID], 
	[ol].[Quantity]
FROM [Sales].[Orders] [o]
JOIN [Sales].[OrderLines] [ol] 
	ON [o].[OrderID] = [ol].[OrderID];
GO


DBCC TRACESTATUS;
GO

/*
	run a few queries and then check the plan
	click on the SELECT operator with the Properties window open, youll see an entry for TraceFlags at the bottom
	can see the three trace flags that I have enabled, and its noted whether they are enabled globally or locally, and at compile time (IsCompileTime True) and run time (IsCompileTime False)
	 go directly to the XML to find the informatio
*/
SELECT [si].[CustomerID], [si].[InvoiceDate], [sil].[StockItemID],
	[sil].[Quantity], [sil].[UnitPrice], ([sil].[Quantity]*[sil].[UnitPrice]) [Total]
FROM [Sales].[Invoices] [si]
JOIN [Sales].[InvoiceLines] [sil]
	ON [si].[InvoiceID] = [sil].[InvoiceID]
WHERE [si].[InvoiceDate] IS NOT NULL
ORDER BY [si].[InvoiceDate] DESC;
GO

/*
	What happens if I use a trace flag, like 9130 to push out residual predicates into a FILTER operator, in a query hint
		see the FILTER operators in the plan:
		nd in the XML we see 9130 for the compilation as a session trace flag, but it doesnt show up in the execution flags
*/
SELECT [si].[CustomerID], [si].[InvoiceDate], [sil].[StockItemID],
	[sil].[Quantity], [sil].[UnitPrice], ([sil].[Quantity]*[sil].[UnitPrice]) [Total]
FROM [Sales].[Invoices] [si]
JOIN [Sales].[InvoiceLines] [sil]
	ON [si].[InvoiceID] = [sil].[InvoiceID]
WHERE [si].[InvoiceDate] IS NOT NULL
ORDER BY [si].[InvoiceDate] DESC
OPTION (QUERYTRACEON 9481);
GO

/*
	Ive run the query in Management Studio and captured the actual plan.  Do I get all the same information from the plan cache?  To check, Ill clear the plan cache, re-run the most query, and then interrogate the plan cache to see what I get:
*/

DBCC FREEPROCCACHE; /* not for production use! */
GO
 
SELECT [si].[CustomerID], [si].[InvoiceDate], [sil].[StockItemID],
	[sil].[Quantity], [sil].[UnitPrice], ([sil].[Quantity]*[sil].[UnitPrice]) [Total]
FROM [Sales].[Invoices] [si]
JOIN [Sales].[InvoiceLines] [sil]
	ON [si].[InvoiceID] = [sil].[InvoiceID]
WHERE [si].[InvoiceDate] IS NOT NULL
ORDER BY [si].[InvoiceDate] DESC
OPTION (QUERYTRACEON 9481);
GO
 
 /*
	open the query plan from the output and look at the XML, I see compilation information, but no runtime info:
*/
SET TRANSACTION ISOLATION LEVEL READ UNCOMMITTED;
GO
 
SELECT [s].[text], [qs].[last_execution_time], [qp].[query_plan]
FROM [sys].[dm_exec_query_stats] [qs]
CROSS APPLY [sys].[dm_exec_query_plan] ([qs].[plan_handle]) qp
CROSS APPLY [sys].[dm_exec_sql_text]([qs].[plan_handle]) [s]
WHERE [s].[text] LIKE '%InvoiceDate%';
GO
 
SET TRANSACTION ISOLATION LEVEL READ COMMITTED;
GO

/*
	if I run the query again, without clearing the plan cache, and taking off the 9130 trace flag, I would get a new entry:
*/
SELECT [si].[CustomerID], [si].[InvoiceDate], [sil].[StockItemID],
	[sil].[Quantity], [sil].[UnitPrice], ([sil].[Quantity]*[sil].[UnitPrice]) [Total]
FROM [Sales].[Invoices] [si]
JOIN [Sales].[InvoiceLines] [sil]
	ON [si].[InvoiceID] = [sil].[InvoiceID]
WHERE [si].[InvoiceDate] IS NOT NULL
ORDER BY [si].[InvoiceDate] DESC;
GO
 
/*
	check the plan for the second entry, with the later execution time, in the XML I find
*/
SET TRANSACTION ISOLATION LEVEL READ UNCOMMITTED;
GO
 
SELECT [s].[text], [qs].[last_execution_time], [qp].[query_plan]
FROM [sys].[dm_exec_query_stats] [qs]
CROSS APPLY [sys].[dm_exec_query_plan] ([qs].[plan_handle]) qp
CROSS APPLY [sys].[dm_exec_sql_text]([qs].[plan_handle]) [s]
WHERE [s].[text] LIKE '%InvoiceDate%';
GO
 
SET TRANSACTION ISOLATION LEVEL READ COMMITTED;
GO


/*
	find in the plan cache
*/
SET TRANSACTION ISOLATION LEVEL READ UNCOMMITTED;
GO
 
WITH XMLNAMESPACES
(DEFAULT 'http://schemas.microsoft.com/sqlserver/2004/07/showplan')
SELECT
STUFF((SELECT ', ' +  tf.value('(./@Value)[1]', 'varchar(10)')
FROM stmt.nodes('./QueryPlan/TraceFlags[@IsCompileTime="1"]/TraceFlag[@Scope="Global"]') AS t(tf)
FOR XML PATH('')
), 1, 2, '') AS GlobalTraceFlags,
STUFF((SELECT ', ' +  tf.value('(./@Value)[1]', 'varchar(10)')
FROM stmt.nodes('./QueryPlan/TraceFlags[@IsCompileTime="1"]/TraceFlag[@Scope="Session"]') AS t(tf)
FOR XML PATH('')
), 1, 2, '') AS SessionTraceFlags,
stmt.query('.') as stmt_node,
cp.usecounts,
qp.query_plan,
cp.plan_handle,
qp.objectid,
cp.objtype,
cp.cacheobjtype,
cp.size_in_bytes
FROM sys.dm_exec_cached_plans AS cp
CROSS APPLY sys.dm_exec_query_plan(plan_handle) AS qp
CROSS APPLY query_plan.nodes('/ShowPlanXML/BatchSequence/Batch/Statements/StmtSimple') AS batch(stmt)
WHERE stmt.exist('./QueryPlan/TraceFlags[@IsCompileTime="1"]') = 1
OPTION(RECOMPILE, MAXDOP 1);



/*
	Clear cache
	*not for production*
*/
DBCC DROPCLEANBUFFERS;

/*
	Clustered Index Seek -> 1 Row
	What's the Estimated I/O cost?
	-- 0.003125
*/
SELECT 
	[FullName], 
	[PhoneNumber]
FROM [Application].[People]
WHERE [PersonID] = 28;


/*
	Find the Page ID so we can check the cache
	NOTE: fn_PhysLocFormatter is undocumented (thus unsupported)
*/
SELECT  
	[FullName], 
	[PhoneNumber], 
	sys.fn_PhysLocFormatter(%%physloc%%)
FROM [Application].[People]
WHERE [PersonID] = 28;

SELECT *
FROM sys.dm_os_buffer_descriptors
WHERE
	[file_id] = 3 AND
	[page_id] = 11 AND
	[database_id] = DB_ID();

/*
	Since it's now cached, does cost change?
*/
SELECT 
	[FullName], 
	[PhoneNumber]
FROM [Application].[People]
WHERE [PersonID] = 28;


/*
	Remove it from cache
*/
DBCC DROPCLEANBUFFERS;


/*
	What about a Clustered Index Scan?
	What's the estimated I/O?
	-- .0609
*/
SELECT 
	[PersonID],
	[FullName], 
	[PhoneNumber]
FROM [Application].[People];

/*
	How many pages in the table?
*/
SELECT 
	[in_row_data_page_count]
FROM [sys].[dm_db_partition_stats]
WHERE [object_id] = OBJECT_ID('Application.People') 
	AND [index_id] = 1;


/*
	Random I/O - 0.003125
	Sequential - 0.000740741
	Compare the output below to the estimated I/O value in the plan
*/
SELECT	0.003125 + (0.000740741 * (79-1));

SELECT COUNT(*) FROM [Application].[People] --1117
	
/*
	Now let's change the statistics for Application.People
	** Don't do this in production, please **
*/
UPDATE STATISTICS [Application].[People]
WITH ROWCOUNT = 1117, PAGECOUNT = 149000;


/*
	What's the new estimated I/O?
	-- 110.373
*/
SELECT 
	[PersonID],
	[FullName], 
	[PhoneNumber]
FROM [Application].[People]
OPTION (RECOMPILE);
GO

/*
	110.373
*/
SELECT	0.003125 +
		0.000740741 * (149000-1)
GO


/*
	reset the numbers
*/
UPDATE STATISTICS [Application].[People]
WITH ROWCOUNT = 1117, PAGECOUNT = 79;
GO


/*
	Inflating the page count while maintaining the row count
	Note original cost again for baseline
*/
SELECT 
	[PersonID],
	[FullName], 
	[PhoneNumber]
FROM [Application].[People]
OPTION (RECOMPILE);
GO


/*
	Inflate row count, but not page count
*/
UPDATE STATISTICS [Application].[People]
WITH ROWCOUNT = 15000000, PAGECOUNT = 79;

/*
 	check page and row count
*/
SELECT 
	[in_row_data_page_count], 
	[row_count]
FROM [sys].[dm_db_partition_stats]
WHERE [object_id] = OBJECT_ID('Application.People') AND [index_id] = 1;


/*
	re-run the query
*/
SELECT 
	[PersonID],
	[FullName], 
	[PhoneNumber]
FROM [Application].[People]
OPTION (RECOMPILE);
GO

/*
	Estimated CPU Cost = 16.5 (increased)
	Estimated I/O Cost = 0.0609 (same as before)
*/


/*
	reset
*/
UPDATE STATISTICS [Application].[People]
WITH ROWCOUNT = 1117, PAGECOUNT = 79;
GO


/*
	Enable Live Query Stats
*/
SELECT [si].[CustomerID], [si].[InvoiceDate], [sil].[StockItemID],
	[sil].[Quantity], [sil].[UnitPrice], ([sil].[Quantity]*[sil].[UnitPrice]) [Total]
FROM [Sales].[Invoices] [si]
JOIN [Sales].[InvoiceLines] [sil]
	ON [si].[InvoiceID] = [sil].[InvoiceID]
WHERE [si].[InvoiceDate] IS NOT NULL
ORDER BY [si].[InvoiceDate] DESC;
GO

/*
	Re-run query and look at live stats through Activity Monitor
*/
SELECT [si].[CustomerID], [si].[InvoiceDate], [sil].[StockItemID],
	[sil].[Quantity], [sil].[UnitPrice], ([sil].[Quantity]*[sil].[UnitPrice]) [Total]
FROM [Sales].[Invoices] [si]
JOIN [Sales].[InvoiceLines] [sil]
	ON [si].[InvoiceID] = [sil].[InvoiceID]
WHERE [si].[InvoiceDate] IS NOT NULL
ORDER BY [si].[InvoiceDate] DESC;
GO


/*
	Copy to another window, change session_id
*/
SELECT *
FROM sys.dm_exec_query_profiles   
WHERE session_id=59;
GO

SELECT node_id,physical_operator_name, SUM(row_count) row_count, 
  SUM(estimate_row_count) AS estimate_row_count, 
  CAST(SUM(row_count)*100 AS float)/SUM(estimate_row_count)  
FROM sys.dm_exec_query_profiles   
WHERE session_id=59
GROUP BY node_id,physical_operator_name  
ORDER BY node_id;  
GO

/*
	Run one more time
*/
SELECT [si].[CustomerID], [si].[InvoiceDate], [sil].[StockItemID],
	[sil].[Quantity], [sil].[UnitPrice], ([sil].[Quantity]*[sil].[UnitPrice]) [Total]
FROM [Sales].[Invoices] [si]
JOIN [Sales].[InvoiceLines] [sil]
	ON [si].[InvoiceID] = [sil].[InvoiceID]
WHERE [si].[InvoiceDate] IS NOT NULL
ORDER BY [si].[InvoiceDate] DESC;
GO


/*
	End
*/
